# -*- coding: utf-8 -*-
# =============================================================================
# Ural URL Extraction Function
# =============================================================================
#
# A function returning an iterator over the urls present in the string argument.
#
import re

from ural.patterns import URL_IN_TEXT_RE

IRRELEVANT_PUNCTUATION = set("!?#\"$%&'()*+,-.:;<=>@[\\]^_`{|}~…’‘`‛«»„‟“”-‐‒–—―−‑⁃,،、")


def urls_from_text(string):
    """
    Function returning an iterator over the urls present in the string argument.

    Args:
        string (str): source string.

    Yields:
        str: an url.

    """
    for match in re.finditer(URL_IN_TEXT_RE, string):
        url = match.group(0)
        s = match.start()

        if s > 0 and string[s - 1] == "[":
            if "](" in url:
                remainder, url = url.split("](", 1)
                yield remainder.strip()

        last_punct = None

        stop = len(url) - 1
        i = stop

        while i != 0 and url[i] in IRRELEVANT_PUNCTUATION and url[i] != last_punct:
            last_punct = url[i]
            i -= 1

        if i != stop:
            url = url[: i + 1]

        yield url



# =============================================================================
# Ural TLD-related Functions
# =============================================================================
#
# Some useful links:
#   - https://wiki.mozilla.org/Public_Suffix_List
#   - https://publicsuffix.org/list/public_suffix_list.dat
#   - https://www.iana.org/domains/root/db
#   - https://data.iana.org/TLD/tlds-alpha-by-domain.txt
#   - http://en.wikipedia.org/wiki/List_of_Internet_top-level_domains
#
# NOTE: there is a debate about what is or is not a TLD per se. This is a very
# interesting topic and one should note that the current module follows an
# "operational" definition of a what a TLD is, but not the purist idea of what
# it actually is actually supposed to be (according to IANA at least). Our
# definition is in line with Mozilla's one wrt domain attribution and the
# problem we are trying to address is usually to obtain the accurate domain name
# from a given url, which is not trivial and means you MUST rely on an exhaustive
# and maintained list of current possibilities (as represented by Mozilla's
# public suffix list).
#
# For instance, let's observe the following url:
#   https://www.whatever.co.uk/section/article.html
#
# What we need to extract from this url is what is colloquially known as a
# domain name: "whatever.co.uk"
#
# So we will consider this url TLD as being: ".co.uk", while a purist would
# argue that the actual TLD is ".uk".
#
import codecs
from os.path import dirname, join

try:
    from urllib.request import urlopen
except ImportError:
    from urllib2 import urlopen

import ural.tld_data as tld_data
from ural.classes.suffix_trie import SuffixTrie
from ural.exceptions import TLDUpgradeError
from ural.utils import attempt_to_decode_idna, safe_urlsplit

MOZILLA_PUBLIC_SUFFIX_LIST = "https://publicsuffix.org/list/public_suffix_list.dat"
IANA_TLD_URL = "https://data.iana.org/TLD/tlds-alpha-by-domain.txt"

SUFFIX_TRIE = SuffixTrie()
TLD_SET = set()


def download(url):
    response = urlopen(url)

    try:
        data = response.read()
    finally:
        response.close()

    return data.decode("utf-8")


def suffix_list_iter(txt):
    in_private_section = False

    for line in txt.split("\n"):

        # Private section
        if "===BEGIN PRIVATE DOMAINS===" in line:
            in_private_section = True

        # Puny code version
        if "// xn--" in line:
            line = line.split()[1]

        line = line.strip()

        # Comments and paragraph breaks
        if not line or line[0] in ("/", "\n"):
            continue

        yield in_private_section, line


def get_suffix_lists(txt):
    public_list = []
    private_list = []

    for private, suffix in suffix_list_iter(txt):
        if private:
            private_list.append(suffix)
        else:
            public_list.append(suffix)

    return public_list, private_list


def parse_tlds(txt):
    tlds = []

    for line in txt.split("\n"):
        line = line.strip()

        if not line or line.startswith("#"):
            continue

        tlds.append(attempt_to_decode_idna(line.lower()))

    return tlds


def refresh():
    global SUFFIX_TRIE
    global TLD_TRIE

    SUFFIX_TRIE = SuffixTrie()

    for suffix in tld_data.PUBLIC_SUFFIXES:
        SUFFIX_TRIE.add(suffix, private=False)

    for suffix in tld_data.PRIVATE_SUFFIXES:
        SUFFIX_TRIE.add(suffix, private=True)

    for tld in tld_data.TLDS:
        TLD_SET.add(tld)


refresh()


def upgrade(transient=False):
    try:
        mozilla_txt = download(MOZILLA_PUBLIC_SUFFIX_LIST)
        iana_txt = download(IANA_TLD_URL)

        output_path = join(dirname(__file__), "tld_data.py")

        public, private = get_suffix_lists(mozilla_txt)
        tlds = parse_tlds(iana_txt)

        tld_data.PUBLIC_SUFFIXES = public
        tld_data.PRIVATE_SUFFIXES = private
        tld_data.TLDS = tlds

        refresh()

        if transient:
            return

        with codecs.open(output_path, "w", encoding="utf-8") as f:
            f.write("# coding: utf-8\n")
            f.write("from __future__ import unicode_literals\n\n")
            f.write("PUBLIC_SUFFIXES = [\n")
            for suffix in public:
                f.write('  "%s",\n' % suffix)
            f.write("]\n\n")
            f.write("PRIVATE_SUFFIXES = [\n")
            for suffix in private:
                f.write('  "%s",\n' % suffix)
            f.write("]\n\n")
            f.write("TLDS = [\n")
            for tld in tlds:
                f.write('  "%s",\n' % tld)
            f.write("]\n")

    except Exception as reason:
        raise TLDUpgradeError("Could not upgrade TLD lists", reason=reason)


def get_domain_name(url):
    """
    Function returning the given url's domain name as a string. This function
    is of course TLD aware.

    Args:
        url (string): Target url.

    Returns:
        str: the url's domain name.

    """
    return SUFFIX_TRIE.extract_domain_name(url)


def has_valid_suffix(url):
    return SUFFIX_TRIE.has_valid_domain_name(url)


def split_suffix(url):
    return SUFFIX_TRIE.split(url)


def is_valid_tld(tld):
    tld = attempt_to_decode_idna(tld.lstrip(".").lower())

    return tld in TLD_SET


def has_valid_tld(url):
    parsed = safe_urlsplit(url)

    if not parsed.hostname:
        return False

    last_part = parsed.hostname.rsplit(".", 1)[-1]

    return is_valid_tld(last_part)



from typing import Iterator
from ural.types import AnyUrlTarget

class HostnameTrieSet(object):
    def __len__(self) -> int: ...
    def add(self, hostname: str) -> None: ...
    def match(self, url: AnyUrlTarget) -> bool: ...
    def __iter__(self) -> Iterator[str]: ...
    def __contains__(self, hostname: str) -> None: ...



# -*- coding: utf-8 -*-
# =============================================================================
# Ural Is Homepage Function
# =============================================================================
#
# Function returning whether the given url looks like a website's homepage.
#
from os.path import splitext

from ural.utils import safe_urlsplit

HOMEPAGE_PATHS = ["", "/", "/index", "/home"]


def is_homepage(url):
    """
    Function returning whether the given url looks like a website's homepage.

    Args:
        url (str): url to test.

    Returns:
        bool

    """
    parsed = safe_urlsplit(url)
    path = parsed.path.strip().rstrip("/")
    path, _ = splitext(path)

    return path in HOMEPAGE_PATHS



from typing import List
from urllib.parse import SplitResult

def lru_stems_from_parsed_url(
    parsed_url: SplitResult, suffix_aware: bool = ...
) -> List[str]: ...
def lru_stems(url: str, suffix_aware: bool = ...): ...
def normalized_lru_stems(url: str, suffix_aware: bool = ..., **kwargs): ...



from setuptools import setup, find_packages

with open("./README.md", "r") as f:
    long_description = f.read()

setup(
    name="ural",
    version="1.0.0-a4",
    description="A helper library full of URL-related heuristics.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="http://github.com/medialab/ural",
    author="Guillaume Plique, Jules Farjas, Oubine Perrin, Benjamin Ooghe-Tabanou, Martin Delabre, Pauline Breteau, Jean Descamps, Béatrice Mazoyer, Amélie Pellé, Laura Miguel, César Pichon",
    author_email="guillaume.plique@sciencespo.fr",
    keywords="url",
    license="GPL-3.0",
    python_requires=">=2.7",
    packages=find_packages(exclude=["scripts", "test"]),
    package_data={"docs": ["README.md"], "ural": ["*.pyi", "**/*.pyi"]},
    install_requires=[],
    extras_require={":python_version<'3.8'": ["typing_extensions"]},
    zip_safe=True,
)



# =============================================================================
# Ural URL Extraction From HTML Function
# =============================================================================
#
# A function returning an iterator over the urls present in the HTML string
# argument.
#
from __future__ import unicode_literals
from ural.patterns import URL_IN_HTML_RE, URL_IN_HTML_BINARY_RE


def __urls_finditer(string):
    for match in URL_IN_HTML_RE.finditer(string):
        url = match.group(1)

        if url is not None:
            url = url.strip('"')
        else:
            url = match.group(2)

            if url is not None:
                url = url.strip("'")
            else:
                url = match.group(3)

        assert url is not None

        yield url


def __urls_finditer_binary(string, encoding="utf-8", errors="strict"):
    for match in URL_IN_HTML_BINARY_RE.finditer(string):
        url = match.group(1)

        if url is not None:
            url = url.strip(b'"')
        else:
            url = match.group(2)

            if url is not None:
                url = url.strip(b"'")
            else:
                url = match.group(3)

        assert url is not None

        yield url.decode(encoding, errors=errors)


def urls_from_html(string, encoding="utf-8", errors="strict"):
    if isinstance(string, bytes):
        return __urls_finditer_binary(string, encoding=encoding, errors=errors)

    return __urls_finditer(string)



# =============================================================================
# Ural LRUTrie Classes
# =============================================================================
#
# Bunch of classes relying on LRUs and prefix trees to perform complex URL
# matching routines.
#
from typing import Generic, TypeVar, Optional, List, Iterator

Metadata = TypeVar("Metadata")

class LRUTrie(Generic[Metadata]):
    def __init__(self, suffix_aware: bool = ...): ...
    def __len__(self) -> int: ...
    def set(self, url: str, metadata: Metadata) -> None: ...
    def __setitem__(self, url: str, metadata: Metadata) -> None: ...
    def match(self, url: str) -> Optional[Metadata]: ...
    def set_lru(self, lru: List[str], metadata: Metadata): ...
    def match_lru(self, lru: List[str]) -> Optional[Metadata]: ...
    def __iter__(self) -> Iterator[Metadata]: ...

class NormalizedLRUTrie(LRUTrie):
    def __init__(self, suffix_aware: bool = ..., **kwargs): ...



# =============================================================================
# Ural Youtube-related heuristic functions
# =============================================================================
#
# Collection of functions related to Youtube urls.
#
import re
from collections import namedtuple

from ural.classes.hostname_trie_set import HostnameTrieSet
from ural.utils import pathsplit, safe_urlsplit
from ural.infer_redirection import infer_redirection
from ural.patterns import QUERY_VALUE_TEMPLATE

YOUTUBE_DOMAINS = [
    "blog.youtube",
    "rewind.youtube",
    "youtu.be",
    "youtube.ae",
    "youtube.al",
    "youtube.am",
    "youtube.at",
    "youtube.az",
    "youtube.ba",
    "youtube.be",
    "youtube.bg",
    "youtube.bh",
    "youtube.bo",
    "youtube.by",
    "youtube.ca",
    "youtube.cat",
    "youtube.ch",
    "youtube.cl",
    "youtube.co",
    "youtube.co.ae",
    "youtube.co.at",
    "youtube.co.cr",
    "youtube.co.hu",
    "youtube.co.id",
    "youtube.co.il",
    "youtube.co.in",
    "youtube.co.jp",
    "youtube.co.ke",
    "youtube.co.kr",
    "youtube.co.ma",
    "youtube.co.nz",
    "youtube.co.th",
    "youtube.co.tz",
    "youtube.co.ug",
    "youtube.co.uk",
    "youtube.co.ve",
    "youtube.co.za",
    "youtube.co.zw",
    "youtube.com",
    "youtube.com.ar",
    "youtube.com.au",
    "youtube.com.az",
    "youtube.com.bd",
    "youtube.com.bh",
    "youtube.com.bo",
    "youtube.com.br",
    "youtube.com.by",
    "youtube.com.co",
    "youtube.com.do",
    "youtube.com.ec",
    "youtube.com.ee",
    "youtube.com.eg",
    "youtube.com.es",
    "youtube.com.gh",
    "youtube.com.gr",
    "youtube.com.gt",
    "youtube.com.hk",
    "youtube.com.hn",
    "youtube.com.hr",
    "youtube.com.jm",
    "youtube.com.jo",
    "youtube.com.kw",
    "youtube.com.lb",
    "youtube.com.lv",
    "youtube.com.ly",
    "youtube.com.mk",
    "youtube.com.mt",
    "youtube.com.mx",
    "youtube.com.my",
    "youtube.com.ng",
    "youtube.com.ni",
    "youtube.com.om",
    "youtube.com.pa",
    "youtube.com.pe",
    "youtube.com.ph",
    "youtube.com.pk",
    "youtube.com.pt",
    "youtube.com.py",
    "youtube.com.qa",
    "youtube.com.ro",
    "youtube.com.sa",
    "youtube.com.sg",
    "youtube.com.sv",
    "youtube.com.tn",
    "youtube.com.tr",
    "youtube.com.tw",
    "youtube.com.ua",
    "youtube.com.uy",
    "youtube.com.ve",
    "youtube.cr",
    "youtube.cz",
    "youtube.de",
    "youtube.dk",
    "youtube.ee",
    "youtube.es",
    "youtube.fi",
    "youtube.fr",
    "youtube.ge",
    "youtube.googleapis.com",
    "youtube.gr",
    "youtube.gt",
    "youtube.hk",
    "youtube.hr",
    "youtube.hu",
    "youtube.ie",
    "youtube.in",
    "youtube.iq",
    "youtube.is",
    "youtube.it",
    "youtube.jo",
    "youtube.jp",
    "youtube.kr",
    "youtube.kz",
    "youtube.la",
    "youtube.lk",
    "youtube.lt",
    "youtube.lu",
    "youtube.lv",
    "youtube.ly",
    "youtube.ma",
    "youtube.md",
    "youtube.me",
    "youtube.mk",
    "youtube.mn",
    "youtube.mx",
    "youtube.my",
    "youtube.ng",
    "youtube.ni",
    "youtube.nl",
    "youtube.no",
    "youtube.pa",
    "youtube.pe",
    "youtube.ph",
    "youtube.pk",
    "youtube.pl",
    "youtube.pr",
    "youtube.pt",
    "youtube.qa",
    "youtube.ro",
    "youtube.rs",
    "youtube.ru",
    "youtube.sa",
    "youtube.se",
    "youtube.sg",
    "youtube.si",
    "youtube.sk",
    "youtube.sn",
    "youtube.soy",
    "youtube.sv",
    "youtube.tn",
    "youtube.tv",
    "youtube.ua",
    "youtube.ug",
    "youtube.uy",
    "youtube.vn",
    "youtubeeducation.com",
    "youtubekids.com",
    "yt.be",
]
YOUTUBE_VIDEO_ID_RE = re.compile(r"^[a-zA-Z0-9_-]{11}$")
QUERY_V_RE = re.compile(QUERY_VALUE_TEMPLATE % r"v", re.I)
NEXT_V_RE = re.compile(r"next=%2Fwatch%3Fv%3D([^%&]+)", re.I)
NESTED_NEXT_V_RE = re.compile(r"next%3D%252Fwatch%253Fv%253D([^%&]+)", re.I)
FRAGMENT_V_RE = re.compile(
    r"^(?:%2F|/)watch(?:%3F|\?)v(?:%3D|=)([a-zA-Z0-9_-]{11})", re.I
)
QUERY_LIST_RE = re.compile(QUERY_VALUE_TEMPLATE % r"list", re.I)

YOUTUBE_VIDEO_URL_TEMPLATE = "https://www.youtube.com/watch?v=%s"
YOUTUBE_USER_URL_TEMPLATE = "https://www.youtube.com/user/%s"
YOUTUBE_CHANNEL_ID_URL_TEMPLATE = "https://www.youtube.com/channel/%s"

# NOTE: it's possible this does not work with all channels...
# Sometimes I think you need 'https://www.youtube.com/%s' instead
# but there is no way to infer this...
YOUTUBE_CHANNEL_NAME_URL_TEMPLATE = "https://www.youtube.com/%s"

YOUTUBE_CHANNEL_NAME_BLACKLIST = {
    "about",
    "account",
    "ads",
    "creators",
    "feed",
    "howyoutubeworks",
    "new",
    "paid_memberships",
    "playlist",
    "reporthistory",
    "results",
    "t",
}

YoutubeVideo = namedtuple("YoutubeVideo", ["id", "playlist"])
YoutubeUser = namedtuple("YoutubeUser", ["id", "name"])
YoutubeChannel = namedtuple("YoutubeChannel", ["id", "name"])


# NOTE: we use a trie to perform efficient queries and so we don't
# need to test every domain/subdomain linearly
YOUTUBE_DOMAINS_TRIE = HostnameTrieSet()

for domain in YOUTUBE_DOMAINS:
    YOUTUBE_DOMAINS_TRIE.add(domain)


def is_youtube_url(url):
    """
    Function returning whether the given url is a valid Youtube url.

    Args:
        url (str): Url to test.

    Returns:
        bool: Whether given url is from Youtube.

    """
    return YOUTUBE_DOMAINS_TRIE.match(url)


def is_youtube_video_id(value):
    return bool(YOUTUBE_VIDEO_ID_RE.match(value))


def parse_youtube_url(url, fix_common_mistakes=True):
    """
    Function parsing the given url and returning either a YoutubeUser,
    YoutubeChannel, YoutubeVideo or None if nothing of information could be
    found.

    Args:
        url (str): Url to parse.
        fix_common_mistakes (bool, optional): Whether to fix common mistakes
            in Youtube urls as you can find them on the web. Defaults to `True`.

    """

    # Inferring redirection
    url = infer_redirection(url)

    # Continuation urls
    m = NEXT_V_RE.search(url) or NESTED_NEXT_V_RE.search(url)

    mlist_query = QUERY_LIST_RE.search(url)
    list_query = mlist_query.group(1) if mlist_query else None

    if m:
        return YoutubeVideo(id=m.group(1), playlist=list_query)

    # Parsing
    parsed = safe_urlsplit(url)

    if not is_youtube_url(parsed):
        return

    _, _, path, query, fragment = parsed

    # youtu.be
    if parsed.hostname.endswith("youtu.be"):

        if path.count("/") > 0:
            v = pathsplit(path)[0]

            if fix_common_mistakes:
                v = v[:11]

            if not is_youtube_video_id(v):
                return

            return YoutubeVideo(id=v, playlist=list_query)

        return

    # Hidden video in fragment
    if fragment:
        mv = FRAGMENT_V_RE.match(fragment)

        if mv:
            v = mv.group(1)

            if not is_youtube_video_id(v):
                return

            return YoutubeVideo(id=v, playlist=list_query)

    # Typical video url
    if path == "/watch":
        mv = QUERY_V_RE.search(query)

        if mv:
            v = mv.group(1)

            if fix_common_mistakes:
                v = v[:11]

            if not is_youtube_video_id(v):
                return

            return YoutubeVideo(id=v, playlist=list_query)

    # Video file
    elif (
        path.startswith("/v/")
        or path.startswith("/video/")
        or path.startswith("/embed/")
    ):
        v = pathsplit(path)[-1]

        if fix_common_mistakes:
            v = v[:11]

        if not is_youtube_video_id(v):
            return

        return YoutubeVideo(id=v, playlist=list_query)

    # Typical user url
    elif path.startswith("/user/"):
        splitted_path = pathsplit(path)

        if len(splitted_path) < 2:
            return None

        user = splitted_path[1]

        return YoutubeUser(id=None, name=user)

    # Channel path?
    elif path.startswith("/c/"):
        # NOTE: there is an edge case here with the "c" channel that does exist
        # which means /c will parse as a channel, but /c/ will not but I don't
        # want to spend too much time on this weirdness.

        splitted_path = pathsplit(path)

        if len(splitted_path) < 2:
            return None

        name = splitted_path[1]

        return YoutubeChannel(id=None, name=name.lstrip("@"))

    elif path.startswith("/channel/"):
        splitted_path = pathsplit(path)

        if len(splitted_path) < 2:
            return None

        cid = splitted_path[1]

        return YoutubeChannel(id=cid, name=None)

    else:
        path = path.rstrip("/")
        if path.count("/") == 1:
            name = path.lstrip("/")

            if name in YOUTUBE_CHANNEL_NAME_BLACKLIST:
                return

            return YoutubeChannel(id=None, name=name.lstrip("@"))


def extract_video_id_from_youtube_url(url):
    parsed = parse_youtube_url(url)

    if parsed is None or not isinstance(parsed, YoutubeVideo):
        return

    return parsed.id


def normalize_youtube_url(url):
    parsed = parse_youtube_url(url)

    if parsed is None:

        # TODO: should we normalize at least www and protocol here?
        return url

    if isinstance(parsed, YoutubeVideo):
        url = YOUTUBE_VIDEO_URL_TEMPLATE % parsed.id

        if parsed.playlist:
            return url + "&list=%s" % parsed.playlist

        return url

    if isinstance(parsed, YoutubeUser):
        return YOUTUBE_USER_URL_TEMPLATE % parsed.name

    if isinstance(parsed, YoutubeChannel):
        if parsed.id is not None:
            return YOUTUBE_CHANNEL_ID_URL_TEMPLATE % parsed.id

        return YOUTUBE_CHANNEL_NAME_URL_TEMPLATE % parsed.name

    raise TypeError("normalize_youtube_url: impossible path reached")



# NOTE: it would be nice to handle quoting recursivity

from __future__ import unicode_literals
from platform import python_version_tuple

PY2 = python_version_tuple()[0] == "2"

import re
from functools import partial

from ural.utils import quote

# NOTE: the following code for `unquote` is adapted from cpython:
# https://github.com/python/cpython/blob/main/Lib/urllib/parse.py
HEX = "0123456789ABCDEFabcdef"
if PY2:
    HEX_TO_BYTE = {(a + b).encode(): chr(int(a + b, 16)) for a in HEX for b in HEX}
else:
    HEX_TO_BYTE = {(a + b).encode(): bytes.fromhex(a + b) for a in HEX for b in HEX}

ASCII_RE = re.compile("([\x00-\x7f]+)")


def _unquote_impl(string, only_printable=False, unsafe=None):
    string = string.encode("utf-8")
    bits = string.split(b"%")
    if len(bits) == 1:
        return string
    res = bytearray(bits[0])
    append = res.extend

    for item in bits[1:]:
        b = HEX_TO_BYTE.get(item[:2])

        if b is not None:
            if only_printable and b < b" ":
                append(b"%")
                append(item)
            elif unsafe is not None and b in unsafe:
                append(b"%")
                append(item)
            else:
                append(b)
                append(item[2:])
        else:
            append(b"%")
            append(item)

    return res


def _generate_unquoted_parts(string, only_printable=False, unsafe=None):
    previous_match_end = 0
    for ascii_match in ASCII_RE.finditer(string):
        start, end = ascii_match.span()
        yield string[previous_match_end:start]  # Non-ASCII
        # The ascii_match[1] group == string[start:end].

        m = ascii_match.group(1)
        c = _unquote_impl(m, only_printable=only_printable, unsafe=unsafe).decode(
            "utf-8", "replace"
        )

        yield c

        previous_match_end = end
    yield string[previous_match_end:]  # Non-ASCII tail


# NOTE: here, unsafe must be a container of bytes
def unquote(string, only_printable=False, unsafe=None, normalize_space=False):
    if "%" not in string:
        return string

    q = "".join(
        _generate_unquoted_parts(string, only_printable=only_printable, unsafe=unsafe)
    )

    if normalize_space:
        q = q.replace(" ", "%20")

    return q


# NOTE: to safely unquote we don't need to replace invalid character because it would
# imply that the parsed url was invalid from the start (except for spaces)

UNSAFE_FOR_AUTH_ITEM = b" @:"
UNSAFE_FOR_PATH = b" ?#"
UNSAFE_FOR_QUERY_ITEM = b" &=#"
UNSAFE_FOR_FRAGMENT = b" "

# NOTE: those method should only be used on parsed urls to canonicalize/normalize.
safely_unquote_auth_item = partial(
    unquote, only_printable=True, normalize_space=True, unsafe=UNSAFE_FOR_AUTH_ITEM
)
safely_unquote_path = partial(
    unquote, only_printable=True, normalize_space=True, unsafe=UNSAFE_FOR_PATH
)
safely_unquote_query_item = partial(
    unquote, only_printable=True, normalize_space=True, unsafe=UNSAFE_FOR_QUERY_ITEM
)
safely_unquote_fragment = partial(
    unquote, only_printable=True, normalize_space=True, unsafe=UNSAFE_FOR_FRAGMENT
)


def safely_unquote_qsl(qsl):
    return [
        (
            safely_unquote_query_item(key),
            safely_unquote_query_item(value) if value is not None else None,
        )
        for key, value in qsl
    ]


QUOTED_SPLIT_RE = re.compile(r"(%[0-9A-Fa-f]{2})")
QUOTED_RE = re.compile(r"^%[0-9A-Fa-f]{2}$")


def safely_quote_iter(string):
    for piece in QUOTED_SPLIT_RE.split(string):
        if QUOTED_RE.match(piece):
            yield piece
        else:
            yield quote(piece)


def safely_quote(string):
    return "".join(safely_quote_iter(string))


def safely_quote_qsl(qsl):
    return [
        (safely_quote(key), safely_quote(value) if value is not None else None)
        for key, value in qsl
    ]


def upper_match(match):
    return match.group(0).upper()


LOWERCASE_QUOTED_RE = re.compile(r"%(?:[0-9A-F][a-f]|[a-f][0-9A-F]|[a-f]{2})")


def upper_quoted(string):
    return LOWERCASE_QUOTED_RE.sub(upper_match, string)


__all__ = [
    "unquote",
    "safely_unquote_auth",
    "safely_unquote_path",
    "safely_unquote_query_item",
    "safely_unquote_fragment",
    "safely_unquote_qsl",
    "safely_quote",
    "safely_quote_qsl",
    "upper_quoted",
]



from typing import NamedTuple, Union, Optional
from ural.types import AnyUrlTarget

def is_telegram_message_id(value: str) -> bool: ...
def is_telegram_url(url: AnyUrlTarget) -> bool: ...
def convert_telegram_url_to_public(url: str) -> str: ...

class TelegramMessage(NamedTuple):
    name: str
    id: str

class TelegramGroup(NamedTuple):
    id: str

class TelegramChannel(NamedTuple):
    name: str

def parse_telegram_url(
    url: str,
) -> Optional[Union[TelegramMessage, TelegramGroup, TelegramChannel]]: ...
def extract_channel_name_from_telegram_url(url: str) -> Optional[str]: ...



# =============================================================================
# Ural Library Endpoint
# =============================================================================
#
from ural.classes.hostname_trie_set import HostnameTrieSet

from ural.canonicalize_url import canonicalize_url
from ural.could_be_html import could_be_html
from ural.could_be_rss import could_be_rss
from ural.ensure_protocol import ensure_protocol
from ural.fingerprint_url import (
    fingerprint_url,
    get_fingerprinted_hostname,
    fingerprint_hostname,
)
from ural.force_protocol import force_protocol
from ural.format_url import format_url, URLFormatter
from ural.get_hostname import get_hostname
from ural.has_special_host import has_special_host, is_special_host
from ural.infer_redirection import infer_redirection
from ural.is_homepage import is_homepage
from ural.is_shortened_url import is_shortened_url
from ural.is_typo_url import is_typo_url
from ural.is_url import is_url
from ural.normalize_url import (
    normalize_url,
    get_normalized_hostname,
    normalize_hostname,
)
from ural.should_follow_href import should_follow_href
from ural.should_resolve import should_resolve
from ural.strip_protocol import strip_protocol
from ural.tld import (
    get_domain_name,
    has_valid_suffix,
    split_suffix,
    is_valid_tld,
    has_valid_tld,
)
from ural.urls_from_text import urls_from_text
from ural.urls_from_html import urls_from_html
from ural.utils import safe_urlsplit, urlpathsplit, pathsplit, add_query_argument



# =============================================================================
# Ural Twitter-related heuristic functions
# =============================================================================
#
# Collection of functions related to Twitter urls.
#
import re
from collections import namedtuple

from ural.patterns import DOMAIN_TEMPLATE
from ural.utils import SplitResult, safe_urlsplit, pathsplit

TWITTER_DOMAINS_RE = re.compile(r"twitter\.com", re.I)
TWITTER_URL_RE = re.compile(DOMAIN_TEMPLATE % r"(?:[^.]+\.)*twitter\.com", re.I)
TWITTER_FRAGMENT_ROUTING_RE = re.compile(r"^!/?")
TWITTER_SCREEN_NAME_BLACKLIST = {
    "explore",
    "home",
    "hashtag",
    "i",
    "messages",
    "notifications",
    "search",
    "settings",
}

TwitterTweet = namedtuple("TwitterTweet", ["user_screen_name", "id"])
TwitterUser = namedtuple("TwitterUser", ["screen_name"])
TwitterList = namedtuple("TwitterList", ["id"])


def is_twitter_url(url):
    """
    Function returning whether the given url is a valid Twitter url.

    Args:
        url (str): Url to test.

    Returns:
        bool: Whether given url is from Youtube.

    """
    if isinstance(url, SplitResult):
        return bool(re.search(TWITTER_DOMAINS_RE, url.hostname))

    return bool(re.match(TWITTER_URL_RE, url))


def normalize_screen_name(username):
    if username in TWITTER_SCREEN_NAME_BLACKLIST:
        return None

    if username.startswith("@"):
        username = username[1:]

    return username.lower()


def parse_twitter_url(url):
    """
    Function returning a parsed result from a given Twitter url.

    Args:
        url (str) : Url from which to extract information if found.

    Returns:
        TwitterTweet (namedtuple containing a user_screen_name and a tweet id) instance if the url is a link to a tweet
        or TwitterUser (namedtuple containing a user_screen_name) instance if the url is a link to a twitter user,
        or TwitterList (namedtuple containing a list id) instance if the url is a link to a twitter list,
        None othewise.

    """

    if not is_twitter_url(url):
        return None

    parsed = safe_urlsplit(url)
    path = pathsplit(parsed.path)

    if path:
        user_screen_name = normalize_screen_name(path[0])

        if user_screen_name is None:
            if path[0] == "i" and path[1] == "lists" and len(path) == 3:
                return TwitterList(id=path[2])
            return None

        if len(path) == 3:
            return TwitterTweet(user_screen_name=user_screen_name, id=path[2])

        return TwitterUser(screen_name=user_screen_name)

    if parsed.fragment.startswith("!"):
        path = re.sub(TWITTER_FRAGMENT_ROUTING_RE, "", parsed.fragment)

        return parse_twitter_url("twitter.com/" + path)

    return None


def extract_screen_name_from_twitter_url(url):
    """
    Function returning the screen_name from a given Twitter url.

    Args:
        url (str) : Url from which we extract the screen_name if found.

    Returns:
        str : screen_name if the url is a valid twitter url, None otherwise.

    """

    parsed_twitter_url = parse_twitter_url(url)

    if isinstance(parsed_twitter_url, TwitterUser):
        return parsed_twitter_url.screen_name

    if isinstance(parsed_twitter_url, TwitterTweet):
        return parsed_twitter_url.user_screen_name

    return None



# =============================================================================
# Ural Domain Name Getter
# =============================================================================
#
# Function returning an url's domain name.
#
from ural.utils import safe_urlsplit
from ural.has_special_host import is_special_host


def get_hostname(url):
    try:
        return safe_urlsplit(url).hostname or None
    except ValueError:
        return None


def get_hostname_prefixes(hostname):
    result = []

    if is_special_host(hostname):
        return [hostname]

    if hostname:
        domain_parts = hostname.split(".")

        for i in range(len(domain_parts)):
            result.append(".".join(domain_parts[i:]))

    return result



# =============================================================================
# Ural URL Normalization Function
# =============================================================================
#
# A handy function relying on heuristics to find and drop irrelevant or
# non-discriminant parts of a URL.
#
import re
from os.path import splitext

from ural.ensure_protocol import ensure_protocol
from ural.infer_redirection import infer_redirection as resolve
from ural.utils import (
    safe_qsl_iter,
    safe_serialize_qsl,
    urlsplit,
    urlunsplit,
    unsplit_netloc,
    decode_punycode_hostname,
    normpath,
    fix_common_query_mistakes,
    SplitResult,
)
from ural.quote import (
    safely_unquote_auth_item,
    safely_unquote_path,
    safely_unquote_qsl,
    safely_unquote_fragment,
    safely_quote,
    safely_quote_qsl,
    upper_quoted,
)
from ural.patterns import PROTOCOL_RE, CONTROL_CHARS_RE

IRRELEVANT_QUERY_PATTERN = r"^(?:__twitter_impression|_guc_consent_skip|guccounter|fb_action_types|(?:php|asp|j)?sessionid|fb_action_ids|fb_source|echobox|feature|recruiter|_unique_id|twclid|mibextid|campaignid|adgroupid|cn-reloaded|ao_noptimize|mkt_tok|fbclid|igshid|refid|gclid|mc_cid|mc_eid|__tn__|_ft_|dclid|wpamp|fref|usqp|ncid|mtm_.+|utm_.+%s|s?een|cftoken|cfid|sid|xt(?:loc|ref|cr|np|or|s)|at_.+|_ga)$"

IRRELEVANT_SUBDOMAIN_PATTERN = r"\b(?:www\d?|mobile%s|m)\."

AMP_QUERY_PATTERN = r"|amp_.+|amp"
AMP_QUERY_COMBOS = {"outputtype": ("amp",)}
AMP_SUBDOMAIN_PATTERN = r"|amp"
AMP_SUFFIXES_RE = re.compile(r"(?:\.amp(?=\.html$)|\.amp/?$|(?<=/)amp/?$)", re.I)

IRRELEVANT_QUERY_RE = re.compile(IRRELEVANT_QUERY_PATTERN % r"", re.I)
IRRELEVANT_SUBDOMAIN_RE = re.compile(IRRELEVANT_SUBDOMAIN_PATTERN % r"", re.I)

IRRELEVANT_QUERY_AMP_RE = re.compile(IRRELEVANT_QUERY_PATTERN % AMP_QUERY_PATTERN, re.I)
IRRELEVANT_SUBDOMAIN_AMP_RE = re.compile(
    IRRELEVANT_SUBDOMAIN_PATTERN % AMP_SUBDOMAIN_PATTERN, re.I
)

IRRELEVANT_QUERY_COMBOS = {
    "marfeeltn": ("amp",),
    "mode": ("amp",),
    "output": ("amp",),
    "platform": ("hootsuite",),
    "fromref": ("twitter",),
    "m": (
        "0",
        "1",
    ),
    "ref": set(
        [
            "bookmark",
            "bookmarks",
            "distributor_share",
            "fb",
            "fb_i",
            "m_notif",
            "nf",
            "notif",
            "shortener",
            "ts",
            "tw",
            "tw_i",
            "twhr",
            "twhs",
            "twitter",
            "viral",
            "feed",
            "twtrec",
        ]
    ),
    "s": lambda v: len(v) <= 2 and all("0" <= x <= "9" for x in v),
    "source": ("twitter",),
    "sns": ("tw",),
    "spref": ("fb", "ts", "tw", "tw_i", "twitter"),
    "_ss": ("r",),
}

# NOTE: if this list becomes too long, switch to HostnameTrieMap
PER_DOMAIN_QUERY_FILTERS = [
    ("facebook.com", lambda k, v: k == "_rdc" or k == "_rdr"),
    (
        "youtube.com",
        lambda k, v: k == "t"
        or k == "si"
        or k == "cbrd"
        or k == "ucbcb"
        or k == "ab_channel",
    ),
]


def qsl_sort_key(item):
    return item[0], item[1] or "", 0 if item[1] is None else 1


def should_strip_query_item(
    item, normalize_amp=True, query_item_filter=None, domain_filter=None
):
    key = item[0].lower()

    pattern = IRRELEVANT_QUERY_AMP_RE if normalize_amp else IRRELEVANT_QUERY_RE

    if pattern.match(key):
        return True

    value = item[1]

    if key in IRRELEVANT_QUERY_COMBOS:
        result = IRRELEVANT_QUERY_COMBOS[key]
        if callable(result):
            return result(value)
        return value in IRRELEVANT_QUERY_COMBOS[key]

    # NOTE: only keep elif because query combos and amp query combos
    # are mutually exclusive.
    elif normalize_amp and key in AMP_QUERY_COMBOS:
        return value in AMP_QUERY_COMBOS[key]

    if domain_filter is not None:
        return domain_filter(key, value)

    if query_item_filter is not None:
        return not query_item_filter(key, value)

    return False


def should_strip_fragment(fragment):
    if fragment == "!/" or fragment == "/" or fragment == "!":
        return False

    return fragment.startswith("/") or fragment.startswith("!")


def normalize_hostname(hostname, normalize_amp=True):
    hostname = hostname.strip().lower()
    hostname = CONTROL_CHARS_RE.sub("", hostname)

    pattern = IRRELEVANT_SUBDOMAIN_AMP_RE if normalize_amp else IRRELEVANT_SUBDOMAIN_RE

    hostname = pattern.sub("", hostname)

    if normalize_amp and hostname.startswith("amp-"):
        hostname = hostname[4:]

    hostname = decode_punycode_hostname(hostname)

    return hostname


def get_normalized_hostname(url, normalize_amp=True, infer_redirection=True):
    if infer_redirection:
        url = resolve(url)

    if isinstance(url, SplitResult):
        splitted = url
    else:
        try:
            splitted = urlsplit(ensure_protocol(url.strip()))
        except ValueError:
            return None

    if not splitted.hostname:
        return None

    return normalize_hostname(splitted.hostname, normalize_amp=normalize_amp)


# NOTE: normalize_url is not suited to be able to process already splitted urls,
# because of mutliple string preprocessing tricks and redirection inferrence
# NOTE: query filter run before quoting shenanigans because targeted pattern
# are not subject to escaping, usually
def normalize_url(
    url,
    sort_query=True,
    strip_authentication=True,
    strip_trailing_slash=True,
    strip_index=True,
    strip_protocol=True,
    strip_irrelevant_subdomains=True,
    strip_fragment="except-routing",
    normalize_amp=True,
    fix_common_mistakes=True,
    infer_redirection=True,
    # NOTE: following arguments currently undocumented
    unsplit=True,
    quoted=False,
    query_item_filter=None,
):
    """
    Function normalizing the given url by stripping it of usually
    non-discriminant parts such as irrelevant query items or sub-domains etc.

    This is a very useful utility when attempting to match similar urls
    written slightly differently when shared on social media etc.

    Args:
        url (str): Target URL as a string.
        sort_query (bool, optional): Whether to sort query items or not.
            Defaults to `True`.
        strip_authentication (bool, optional): Whether to drop authentication.
            Defaults to `True`.
        strip_trailing_slash (bool, optional): Whether to drop trailing slash.
            Defaults to `False`.
        strip_index (bool, optional): Whether to drop trailing index at the end
            of the url. Defaults to `True`.
        strip_irrelevant_subdomains (bool, optional): Whether to strip irrelevant subdomains such as www etc.
            Default to True.
        strip_protocol (bool, optional): Whether to strip the url's protocol.
            Defaults to `True`.
        strip_fragment (bool|str, optional): Whether to drop non-routing fragment from the url?
            If set to `except-routing` will only drop non-routing fragment (i.e. fragments that
            do not contain a "/").
            Defaults to `except-routing`.
        normalize_amp (bool, optional): Whether to attempt to normalize Google
            AMP urls. Defaults to True.
        fix_common_mistakes (bool, optional): Whether to attempt solving common mistakes.
            Defaults to True.
        infer_redirection (bool, optional): Whether to attempt resolving common
            redirects by leveraging well-known GET parameters. Defaults to `False`.

    Returns:
        string: The normalized url.

    """
    original_url_arg = url

    if infer_redirection:
        url = resolve(url)

    url = CONTROL_CHARS_RE.sub("", url)
    url = url.strip()
    url = upper_quoted(url)

    has_protocol = PROTOCOL_RE.match(url)

    # Ensuring scheme so parsing works correctly
    if not has_protocol:
        url = "http://" + url

    # Parsing
    try:
        splitted = urlsplit(url)
    except ValueError:
        return original_url_arg

    scheme, netloc, path, query, fragment = splitted
    user, password, hostname, port = (
        splitted.username,
        splitted.password,
        splitted.hostname,
        splitted.port,
    )

    # Fixing common mistakes
    if fix_common_mistakes and query:
        query = fix_common_query_mistakes(query)

    # Handling punycode
    if hostname:
        hostname = decode_punycode_hostname(hostname)

    # Dropping :80 & :443
    if port == 80 or port == 443:
        port = None

    # Normalizing the path
    if path:
        trailing_slash = False
        if path.endswith("/") and len(path) > 1:
            trailing_slash = True
        path = normpath(path)
        if trailing_slash and not strip_trailing_slash:
            path = path + "/"

    # Handling Google AMP suffixes
    if normalize_amp:
        path = AMP_SUFFIXES_RE.sub("", path)

    # Dropping index:
    if strip_index:
        segments = path.rsplit("/", 1)

        if len(segments) != 0:
            last_segment = segments[-1]
            filename, _ = splitext(last_segment)

            if filename == "index" or filename == "default":
                segments.pop()
                path = "/".join(segments)

    # Dropping irrelevant query items
    qsl = []

    if query:
        domain_filter = None

        if splitted.hostname:
            domain_filter = next(
                (
                    f
                    for d, f in PER_DOMAIN_QUERY_FILTERS
                    if splitted.hostname.endswith(d)
                ),
                None,
            )

        # TODO: what to do of empty query items vs. no valued
        # TODO: should be dedupe query items?
        qsl = [
            item
            for item in safe_qsl_iter(query)
            if not should_strip_query_item(
                item,
                normalize_amp=normalize_amp,
                query_item_filter=query_item_filter,
                domain_filter=domain_filter,
            )
        ]

        if sort_query:
            qsl = sorted(qsl, key=qsl_sort_key)

    # Dropping fragment if it's not routing
    if fragment and strip_fragment:
        if strip_fragment is True or not should_strip_fragment(fragment):
            fragment = ""

    # Always dropping trailing slash with empty query & fragment
    if path == "/" and not fragment and not query:
        path = ""

    # Dropping irrelevant subdomains
    if hostname and strip_irrelevant_subdomains:
        hostname = re.sub(
            IRRELEVANT_SUBDOMAIN_AMP_RE if normalize_amp else IRRELEVANT_SUBDOMAIN_RE,
            "",
            hostname,
        )

    # Dropping scheme
    if strip_protocol or not has_protocol:
        scheme = ""

    # Dropping authentication
    if strip_authentication:
        user = None
        password = None

    # Normalizing AMP subdomains
    if normalize_amp and hostname and hostname.startswith("amp-"):
        hostname = hostname[4:]

    # Dropping trailing slash
    if strip_trailing_slash and path.endswith("/"):
        path = path.rstrip("/")

    # Quoting
    if user:
        if quoted:
            user = safely_quote(user)
        else:
            user = safely_unquote_auth_item(user)

    if password:
        if quoted:
            password = safely_quote(password)
        else:
            password = safely_unquote_auth_item(password)

    if quoted:
        path = safely_quote(path)
    else:
        path = safely_unquote_path(path)

    if quoted:
        qsl = safely_quote_qsl(qsl)
    else:
        qsl = safely_unquote_qsl(qsl)

    query = safe_serialize_qsl(qsl)

    if quoted:
        fragment = safely_quote(fragment)
    else:
        fragment = safely_unquote_fragment(fragment)

    # Result
    netloc = unsplit_netloc(user, password, hostname, port)
    result = SplitResult(scheme, netloc.lower(), path, query, fragment)

    if not unsplit:
        return result

    # TODO: check if works with `unsplit=False`
    if strip_protocol or not has_protocol:
        result = urlunsplit(result)[2:]
    else:
        result = urlunsplit(result)

    return result



# =============================================================================
# TrieDict
# =============================================================================
#
# A simple Python implementation of a (prefix, value) Trie.
#
from typing import TypeVar, Generic, Union, Optional, Iterator, Tuple

Value = TypeVar("Value")
DefaultValue = TypeVar("DefaultValue")

class TrieDict(Generic[Value]):
    def __len__(self) -> int: ...
    def __setitem__(self, prefix: str, value: Value) -> None: ...
    def get(
        self, prefix: str, default: DefaultValue = ...
    ) -> Union[Value, DefaultValue]: ...
    def __getitem__(self, prefix: str) -> Value: ...
    def longest_matching_prefix_value(self, prefix: str) -> Optional[Value]: ...
    def items(self) -> Iterator[Tuple[str, Value]]: ...
    def prefixes(self) -> Iterator[str]: ...
    def values(self) -> Iterator[Value]: ...
    def __iter__(self) -> Iterator[Tuple[str, Value]]: ...
    def set_and_prune_if_shorter(self, prefix: str, value: Value) -> None: ...



# =============================================================================
# Ural LRU Stems Function
# =============================================================================
#
# A function returning the url parts in hierarchical order.
#
import re

from ural.utils import urlsplit
from ural.tld import split_suffix
from ural.ensure_protocol import ensure_protocol
from ural.normalize_url import normalize_url
from ural.has_special_host import is_special_host

PORT_SPLITTER = re.compile(r":(?![\d:]+])")


def lru_stems_from_parsed_url(parsed_url, suffix_aware=True):
    scheme, netloc, path, query, fragment = parsed_url
    lru = []

    if scheme:
        lru.append("s:" + scheme)

    user = None
    password = None

    # Handling auth
    if "@" in netloc:
        auth, netloc = netloc.split("@", 1)

        if ":" in auth:
            user, password = auth.split(":", 1)
        else:
            user = auth

    # Parsing domain & port
    netloc = PORT_SPLITTER.split(netloc)

    if len(netloc) == 2:
        port = netloc[1]
        lru.append("t:" + port)

    # Need to process TLD?
    should_process_normally = not suffix_aware

    if suffix_aware:
        split_result = split_suffix(parsed_url)

        if split_result is None:
            should_process_normally = True

        else:
            domain, suffix = split_result
            lru.append("h:" + suffix)

            if domain:
                for element in reversed(domain.split(".")):
                    lru.append("h:" + element)

    if should_process_normally:
        if is_special_host(netloc[0]):
            lru.append("h:" + netloc[0])
        else:
            for element in reversed(netloc[0].split(".")):
                lru.append("h:" + element)

    # Path
    for element in path.split("/")[1:]:
        lru.append("p:" + element)

    # Query
    if query and query[0]:
        lru.append("q:" + query)

    # Fragment
    if fragment and fragment[0]:
        lru.append("f:" + fragment)

    # User
    if user:
        lru.append("u:" + user)

    # Password
    if password:
        lru.append("w:" + password)
    return lru


# TODO: ensure_protocol
def lru_stems(url, suffix_aware=False):
    """
    Function returning the parts of the given url in the hierarchical order (lru).

    Args:
        url (str): Target URL as a string.

    Returns:
        list: The lru, with a prefix identifying the type of each part.
    """

    full_url = ensure_protocol(url)
    return lru_stems_from_parsed_url(urlsplit(full_url), suffix_aware=suffix_aware)


def normalized_lru_stems(url, suffix_aware=False, **kwargs):
    full_url = ensure_protocol(url)
    parsed_url = normalize_url(full_url, unsplit=False, **kwargs)
    return lru_stems_from_parsed_url(parsed_url, suffix_aware=suffix_aware)
